import logging

import dramatiq
from django.apps import apps
from django.db import OperationalError, transaction

from ..conf import settings
from ..contrib.reversion import with_reversion

logger = logging.getLogger(__name__)


def task_runner(*, task_pk, workflow_pk, countdown=None, eta=None, retries=0):
    """Schedule asynchronous machine task using celery."""
    _dramatiq_task_runner.send_with_options(
        args=(task_pk, workflow_pk),
        delay=countdown,
        retries=retries,
    )


class RetryError(Exception):
    """Raised to retry a task if the task result is ``False``."""

    pass


@dramatiq.actor(
    queue_name=settings.JOEFLOW_CELERY_QUEUE_NAME,
    retry_when=lambda a, b: isinstance(b, (OperationalError, RetryError)),
)
def _dramatiq_task_runner(task_pk, workflow_pk, retries=0):
    Task = apps.get_model("joeflow", "Task")
    with transaction.atomic():
        task = Task.objects.select_for_update().get(pk=task_pk, completed=None)

        workflow = (
            task.content_type.model_class()
            .objects.select_for_update(nowait=True)
            .get(pk=workflow_pk)
        )

        try:
            logger.info("Executing %r", task)
            node = task.node
            with_task = getattr(node, "with_task", False)
            kwargs = {}
            if with_task:
                kwargs["task"] = task
            with with_reversion(task):
                result = node(workflow, **kwargs)
        except OperationalError:
            raise
        except:  # NoQA
            task.fail()
            logger.exception("Execution of %r failed", task)
        else:
            if result is False:
                logger.info("Task returned False, retrying …")
                raise RetryError("Task returned False, retrying …")
            elif result is True:
                result = None
            logger.info("Task completed successful, starting next tasks: %s", result)
            task.start_next_tasks(next_nodes=result)
            task.finish()



from appconf import AppConf
from django.conf import settings

__all__ = ("settings",)


class JoeflowAppConfig(AppConf):
    """
    List of available settings.

    To change the default values just set the setting in your settings file.
    """

    JOEFLOW_TASK_RUNNER = "joeflow.runner.dramatiq.task_runner"
    """
    Task runner is used to execute machine tasks.

    JoeFlow supports two different asynchronous task runners – Dramatiq_ and Celery_.

    To use either of the task runners change this setting to:

    * ``joeflow.runner.dramatiq.task_runner``
    * ``joeflow.runner.celery.task_runner``

    .. _Dramatiq: https://dramatiq.io/
    .. _Celery: http://www.celeryproject.org/
    """

    JOEFLOW_CELERY_QUEUE_NAME = "joeflow"
    """
    Queue name in which all machine tasks will be queued.
    """



"""Set of reusable machine tasks."""
from typing import Iterable

from django.utils import timezone

__all__ = (
    "Start",
    "Join",
    "Wait",
)

try:
    # We need to import the Dramatiq task,to ensure
    # it is deteced by django-dramatiq rundramatiq command.
    from joeflow.runner.dramatiq import _dramatiq_task_runner  # NoQA
except ImportError:
    pass


class Start:
    """
    Start a new function via a callable.

    Creates a new workflow instance and executes a start task.
    The start task does not do anything beyond creating the workflow.

    Sample:

    .. code-block:: python

        from django.db import models
        from joeflow.models import Workflow
        from joeflow import tasks


        class StartWorkflow(Workflow):
            a_text_field = models.TextField()

            start = tasks.Start()

            def end(self):
                pass

            edges = (
                (start, end),
            )

        workflow = StartWorkflow.start(a_text_field="initial data")

    """

    def __call__(self, **kwargs):
        workflow = self.workflow_cls.objects.create(**kwargs)
        task = workflow.task_set.create(name=self.name, workflow=workflow)
        task.start_next_tasks()
        return workflow


class Join:
    """
    Wait for all parent tasks to complete before continuing the workflow.

    Args:
        *parents (str): List of parent task names to wait for.

    Sample:

    .. code-block:: python

        from django.db import models
        from joeflow.models import Workflow
        from joeflow import tasks


        class SplitJoinWorkflow(Workflow):
            parallel_task_value = models.PositiveIntegerField(default=0)

            start = tasks.Start()

            def split(self):
                return [self.batman, self.robin]

            def batman(self):
                self.parallel_task_value += 1
                self.save(update_fields=['parallel_task_value'])

            def robin(self):
                self.parallel_task_value += 1
                self.save(update_fields=['parallel_task_value'])

            join = tasks.Join('batman', 'robin')

            edges = (
                (start, split),
                (split, batman),
                (split, robin),
                (batman, join),
                (robin, join),
            )

    """

    with_task = True

    def __init__(self, *parents: Iterable[str]):
        self.parents = set(parents)

    def __call__(self, workflow, task):
        return set(task.parent_task_set.values_list("name", flat=True)) == self.parents

    def create_task(self, workflow, prev_task):
        return workflow.task_set.get_or_create(
            name=self.name,
            type=self.type,
            completed=None,
        )[0]


class Wait:
    """
    Wait for a certain amount of time and then continue with the next tasks.

    Args:
        duration (datetime.timedelta): Time to wait in time delta from creation of task.

    Sample:

    .. code-block:: python

        import datetime

        from django.db import models
        from joeflow.models import Workflow
        from joeflow import tasks


        class WaitWorkflow(Workflow):
            parallel_task_value = models.PositiveIntegerField(default=0)

            start = tasks.Start()

            wait = tasks.Wait(datetime.timedelta(hours=3))

            def end(self):
                pass

            edges = (
                (start, wait),
                (wait, end),
            )

    """

    with_task = True

    def __init__(self, duration: timezone.timedelta):
        self.duration = duration

    def __call__(self, workflow, task):
        return timezone.now() - task.created >= self.duration



"""
A task defines the behavior or a workflow.

A task can be considered as a simple transaction that changes state of a workflow.
There are two types of tasks, human and machine tasks.
"""
from .human import *  # NoQA
from .machine import *  # NoQA

HUMAN = "human"
MACHINE = "machine"



from contextlib import contextmanager

from joeflow import utils

__all__ = (
    "register_workflows",
    "RevisionMixin",
    "with_reversion",
    "VersionAdmin",
)


def register_workflows():
    try:
        from reversion import revisions
    except ImportError:
        pass
    else:
        for workflow in utils.get_workflows():
            if not revisions.is_registered(workflow):
                revisions.register(workflow)


@contextmanager
def with_reversion(task):
    try:
        import reversion
    except ImportError:
        yield
    else:
        with reversion.create_revision():
            yield
            reversion.set_comment(task.name)


try:
    import reversion
    from reversion.views import RevisionMixin as _RevisionMixin

    class RevisionMixin(_RevisionMixin):
        def dispatch(self, request, *args, **kwargs):
            if self.revision_request_creates_revision(request):
                reversion.set_comment(self.name)
            return super().dispatch(request, *args, **kwargs)

except ImportError:

    class RevisionMixin:
        pass


try:
    from reversion.admin import VersionAdmin
except ImportError:
    from django.contrib.admin import ModelAdmin as VersionAdmin



from django.contrib.auth.mixins import PermissionRequiredMixin
from django.db import transaction
from django.forms import modelform_factory
from django.shortcuts import get_object_or_404
from django.views import generic

from . import forms, models
from .contrib.reversion import RevisionMixin


class WorkflowTemplateNameViewMixin:
    name = None

    def get_template_names(self):
        names = [
            "%s/%s_%s.html"
            % (
                self.model._meta.app_label,
                self.model._meta.model_name,
                self.name,
            )
        ]
        names.extend(super().get_template_names())
        names.append(
            "%s/workflow%s.html"
            % (self.model._meta.app_label, self.template_name_suffix)
        )
        return names


class TaskViewMixin(WorkflowTemplateNameViewMixin, RevisionMixin):
    name = None
    path = ""
    """
    URL pattern postfix for task view.

    Should a task require custom arguments via URL, path
    can be set to provide a pattern postfix. e.G.::

        start = tasks.StartView(path="path/to/<other_pk>")

    """

    def __init__(self, **kwargs):
        self._instance_kwargs = kwargs
        super().__init__(**kwargs)

    def get_task(self):
        try:
            return get_object_or_404(
                models.Task,
                pk=self.kwargs["pk"],
                name=self.name,
                completed=None,
            )
        except KeyError:
            return models.Task(name=self.name, type=models.Task.HUMAN)

    def next_task(self):
        task = self.get_task()
        task.workflow = self.model._base_manager.get(pk=self.object.pk)
        task.finish(self.request.user)
        task.start_next_tasks()

    def get_object(self, queryset=None):
        task = self.get_task()
        return task.workflow

    @transaction.atomic
    def form_valid(self, form):
        response = super().form_valid(form)
        self.next_task()
        return response

    def create_task(self, workflow, prev_task):
        """
        Return a new database instance of this task.

        The factory method should be overridden, to create custom database instances
        based on the task node's class, the workflow or the previous task.

        Note:
            This is not a view method, a request via ``self.request`` is not available.

        Args:
            workflow (joeflow.models.Workflow): Current workflow instance.
            prev_task (joeflow.models.Task): Instance of the previous Task.

        Returns:
            joeflow.models.Task: New task instance.

        """
        return workflow.task_set.create(
            name=self.name, type=self.type, workflow=workflow
        )


class WorkflowDetailView(WorkflowTemplateNameViewMixin, generic.DetailView):
    pass


class OverrideView(
    PermissionRequiredMixin,
    RevisionMixin,
    WorkflowTemplateNameViewMixin,
    generic.UpdateView,
):
    permission_required = "override"
    name = "override"
    form_class = forms.OverrideForm
    fields = "__all__"

    def get_form_class(self):
        return modelform_factory(self.model, form=self.form_class, fields=self.fields)

    @transaction.atomic()
    def form_valid(self, form):
        response = super().form_valid(form)
        form.start_next_tasks(self.request.user)
        return response



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as t


class JoeflowConfig(AppConfig):
    name = "joeflow"
    verbose_name = t("Joeflow")

    def ready(self):
        from .contrib.reversion import register_workflows

        register_workflows()



from django import forms
from django.utils.translation import gettext_lazy as t

from . import models


class OverrideForm(forms.ModelForm):
    next_tasks = forms.MultipleChoiceField(
        label=t("Next tasks"),
        choices=[],
        required=False,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["next_tasks"].choices = [
            (name, name) for name in dict(self._meta.model.get_nodes()).keys()
        ]

    def get_next_task_nodes(self):
        names = self.cleaned_data["next_tasks"]
        for name in names:
            yield self._meta.model.get_node(name)

    def start_next_tasks(self, user=None):
        active_tasks = list(self.instance.task_set.filter(completed=None))
        for task in active_tasks:
            task.cancel(user)
        if active_tasks:
            parent_tasks = active_tasks
        else:
            try:
                parent_tasks = [self.instance.task_set.latest()]
            except models.Task.DoesNotExist:
                parent_tasks = []
        override_task = self.instance.task_set.create(
            name="override",
            type=models.Task.HUMAN,
        )
        override_task.parent_task_set.set(parent_tasks)
        override_task.finish(user=user)
        override_task.start_next_tasks(next_nodes=self.get_next_task_nodes())



import types
from collections import defaultdict

import graphviz as gv


def get_workflows() -> types.GeneratorType:
    """Return all registered workflows."""
    from django.apps import apps

    from .models import Workflow

    apps.check_models_ready()
    for model in apps.get_models():
        if issubclass(model, Workflow) and model is not Workflow and model.edges:
            yield model


def get_workflow(name):
    for workflow_cls in get_workflows():
        if (
            name.lower()
            == f"{workflow_cls._meta.app_label}.{workflow_cls.__name__}".lower()
        ):
            return workflow_cls


class NoDashDiGraph(gv.Digraph):
    """
    Like `.graphviz.Digraph` but with unique nodes and edges.

    Nodes and edges are unique and their attributes will be overridden
    should the same node or edge be added twice. Nodes are unique by name
    and edges unique by head and tail.

    Underscores are replaced with whitespaces from identifiers.
    """

    def __init__(self, *args, **kwargs):
        self._nodes = defaultdict(dict)
        self._edges = defaultdict(dict)
        super().__init__(*args, **kwargs)

    def __iter__(self, subgraph=False):
        """Yield the DOT source code line by line (as graph or subgraph)."""
        if self.comment:
            yield self._comment(self.comment)

        if subgraph:
            if self.strict:
                raise ValueError("subgraphs cannot be strict")
            head = self._subgraph if self.name else self._subgraph_plain
        else:
            head = self._head_strict if self.strict else self._head
        yield head(self._quote(self.name) + " " if self.name else "")

        for kw in ("graph", "node", "edge"):
            attrs = getattr(self, "%s_attr" % kw)
            if attrs:
                yield self._attr(kw, self._attr_list(None, attrs))

        yield from self.body

        for name, attrs in sorted(self._nodes.items()):
            name = self._quote(name)
            label = attrs.pop("label", None)
            _attributes = attrs.pop("_attributes", None)
            attr_list = self._attr_list(label, attrs, _attributes)
            yield self._node(name, attr_list)

        for edge, attrs in sorted(self._edges.items()):
            tail_name, head_name = edge
            tail_name = self._quote_edge(tail_name)
            head_name = self._quote_edge(head_name)
            label = attrs.pop("label", None)
            _attributes = attrs.pop("_attributes", None)
            attr_list = self._attr_list(label, attrs, _attributes)
            yield self._edge(tail=tail_name, head=head_name, attr=attr_list)

        yield self._tail

    def node(self, name, **attrs):
        self._nodes[name] = attrs

    def edge(self, tail_name, head_name, **attrs):
        self._edges[(tail_name, head_name)] = attrs

    @staticmethod
    def _quote(identifier, *args, **kwargs):
        """Remove underscores from labels."""
        identifier = identifier.replace("_", " ")
        return gv.quoting.quote(identifier, *args, **kwargs)

    @staticmethod
    def _quote_edge(identifier):
        """Remove underscores from labels."""
        identifier = identifier.replace("_", " ")
        return gv.quoting.quote_edge(identifier)



"""Set of reusable human tasks."""
from django.views import generic

from joeflow.views import TaskViewMixin

__all__ = (
    "StartView",
    "UpdateView",
)


class StartView(TaskViewMixin, generic.CreateView):
    """
    Start a new workflow by a human with a view.

    Starting a workflow with a view allows users to provide initial data.

    Similar to Django's :class:`CreateView<django.views.generic.edit.CreateView>`
    but does not only create the workflow but also completes a tasks.
    """

    pass


class UpdateView(TaskViewMixin, generic.UpdateView):
    """
    Modify the workflow state and complete a human task.

    Similar to Django's :class:`UpdateView<django.views.generic.edit.UpdateView>`
    but does not only update the workflow but also completes a tasks.
    """

    pass



from django.core.management import BaseCommand

from joeflow import utils
from joeflow.models import Workflow


class Command(BaseCommand):
    help = "Render workflow graph to file."

    def add_arguments(self, parser):
        parser.add_argument(
            "workflow",
            nargs="*",
            type=str,
            help="List of workflows to render in the form app_label.workflow_name",
        )
        parser.add_argument(
            "-f",
            "--format",
            dest="format",
            type=str,
            choices=("svg", "pdf", "png"),
            default="svg",
            help="Output file format. Default: svg",
        )
        parser.add_argument(
            "-d",
            "--directory",
            dest="directory",
            type=str,
            help="Output directory. Default is current working directory.",
        )
        parser.add_argument(
            "-c",
            "--cleanup",
            dest="cleanup",
            action="store_true",
            help="Remove dot-files after rendering.",
        )

    def handle(self, *args, **options):
        workflows = options["workflow"]
        verbosity = options["verbosity"]
        file_format = options["format"]
        cleanup = options["cleanup"]
        directory = options.get("directory", None)

        workflows = [utils.get_workflow(s) for s in workflows] or utils.get_workflows()

        for workflow in filter(None, workflows):
            if workflow != Workflow:
                opt = workflow._meta
                if verbosity > 0:
                    self.stdout.write(
                        "Rendering graph for '%s.%s'… "
                        % (opt.app_label, opt.model_name),
                        ending="",
                    )
                filename = f"{opt.app_label}_{workflow.__name__}".lower()
                graph = workflow.get_graph()
                graph.format = file_format
                graph.render(filename=filename, directory=directory, cleanup=cleanup)
                if verbosity > 0:
                    self.stdout.write("Done!", self.style.SUCCESS)
            else:
                self.stderr.write(
                    "%r is not a Workflow subclass" % workflow, self.style.WARNING
                )



import logging
import sys
import traceback
from typing import Any, List, Tuple

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.db import models, transaction
from django.db.models.base import ModelBase
from django.db.models.functions import Now
from django.urls import NoReverseMatch, path, reverse
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.safestring import SafeString
from django.utils.translation import gettext_lazy as t
from django.views.generic.edit import BaseCreateView

from . import tasks, utils, views
from .conf import settings
from .utils import NoDashDiGraph

logger = logging.getLogger(__name__)

__all__ = (
    "Workflow",
    "Task",
)


class WorkflowBase(ModelBase):
    """Set node names on the nodes."""

    def __new__(mcs, name, bases, attrs):
        edges = attrs.get("edges") or []
        klass = super().__new__(mcs, name, bases, attrs)
        nodes = set()
        for edge in edges:
            nodes |= set(edge)

        for name, func in attrs.items():
            try:
                if func in nodes:
                    node = getattr(klass, name)
                    node.name = name
                    node.type = (
                        tasks.HUMAN
                        if isinstance(node, views.TaskViewMixin)
                        else tasks.MACHINE
                    )
                    node.workflow_cls = klass
            except TypeError:
                pass
        return klass


class Workflow(models.Model, metaclass=WorkflowBase):
    """
    The `WorkflowState` object holds the state of a workflow instances.

    It is represented by a Django Model. This way all workflow states
    are persisted in your database.
    """

    id = models.BigAutoField(primary_key=True, editable=False)
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    modified = models.DateTimeField(auto_now=True, db_index=True)

    task_set = GenericRelation(
        "joeflow.Task", object_id_field="_workflow_id", for_concrete_model=False
    )

    class Meta:
        permissions = [("override", t("Can override a workflow."))]

    def save(self, **kwargs):
        if self.pk:
            try:
                update_fields = kwargs["update_fields"]
            except KeyError:
                pass
            else:
                update_fields.append("modified")
        super().save(**kwargs)

    edges: List[Tuple[Any, Any]] = None
    """
    Edges define the transitions between tasks.

    They are the glue that binds tasks together. Edges have no
    behavior but define the structure of a workflow.

    Returns:
        (list[tuple]):
            List of edges. An edge is represented by a tuple
            including start end end not of an edge.
    """

    override_view = views.OverrideView
    detail_view = views.WorkflowDetailView

    @classmethod
    def _wrap_view_instance(cls, name, view_instance):
        return type(view_instance).as_view(
            model=cls, name=name, **view_instance._instance_kwargs
        )

    @classmethod
    def get_nodes(cls):
        nodes = set()
        for edge in cls.edges:
            nodes |= set(edge)
        for node in nodes:
            yield node.name, node

    @classmethod
    def urls(cls):
        """
        Return all URLs to workflow related task and other special views.

        Example::

            from django.urls import path, include

            from . import models

            urlpatterns = [
                # …
                path('myworkflow/', include(models.MyWorkflow.urls())),
            ]

        Returns:
            tuple(list, str): Tuple containing aw list of URLs and the workflow namespace.

        """
        urls = []
        for name, node in cls.get_nodes():
            if isinstance(node, views.TaskViewMixin):
                if isinstance(node, BaseCreateView):
                    route = "{name}/".format(name=name)
                else:
                    route = "{name}/<int:pk>/".format(name=name)
                urls.append(
                    path(
                        route + node.path,
                        cls._wrap_view_instance(name, node),
                        name=name,
                    )
                )
        if cls.detail_view:
            urls.append(
                path("<int:pk>/", cls.detail_view.as_view(model=cls), name="detail")
            )
        if cls.override_view:
            urls.append(
                path(
                    "<int:pk>/override",
                    cls.override_view.as_view(model=cls),
                    name="override",
                )
            )
        return urls, cls.get_url_namespace()

    @classmethod
    def get_node(cls, name: str):
        """Get node by name."""
        return dict(cls.get_nodes())[name]

    @classmethod
    def get_next_nodes(cls, prev_node):
        for start, end in cls.edges:
            if start.name == prev_node.name:
                yield end

    @classmethod
    def get_url_namespace(cls):
        return cls.__name__.lower()

    def get_absolute_url(self):
        """Return URL to workflow detail view."""
        return reverse(
            "{}:detail".format(self.get_url_namespace()), kwargs=dict(pk=self.pk)
        )

    def get_override_url(self):
        """Return URL to workflow override view."""
        return reverse(
            "{}:override".format(self.get_url_namespace()), kwargs=dict(pk=self.pk)
        )

    @classmethod
    def get_graph(cls, color="black"):
        """
        Return workflow graph.

        Returns:
            (graphviz.Digraph): Directed graph of the workflow.

        """
        graph = NoDashDiGraph()
        graph.attr("graph", rankdir="LR")
        graph.attr(
            "node",
            dict(
                fontname="sans-serif", shape="rect", style="filled", fillcolor="white"
            ),
        )
        for name, node in cls.get_nodes():
            node_style = "filled"
            if node.type == tasks.HUMAN:
                node_style += ", rounded"
            graph.node(name, style=node_style, color=color, fontcolor=color)

        for start, end in cls.edges:
            graph.edge(start.name, end.name, color=color)
        return graph

    @classmethod
    def get_graph_svg(cls):
        """
        Return graph representation of a model workflow as SVG.

        The SVG is HTML safe and can be included in a template, e.g.:

        .. code-block:: html

            <html>
            <body>
            <!--// other content //-->
            {{ workflow_class.get_graph_svg }}
            <!--// other content //-->
            </body>
            </html>

        Returns:
            (django.utils.safestring.SafeString): SVG representation of a running workflow.

        """
        graph = cls.get_graph()
        graph.format = "svg"
        return SafeString(graph.pipe(encoding="utf-8"))  # nosec

    get_graph_svg.short_description = t("graph")

    def get_instance_graph(self):
        """Return workflow instance graph."""
        graph = self.get_graph(color="#888888")

        names = dict(self.get_nodes()).keys()

        for task in self.task_set.filter(name__in=names):
            href = task.get_absolute_url()
            style = "filled"
            peripheries = "1"

            if task.type == tasks.HUMAN:
                style += ", rounded"
            if not task.completed:
                style += ", bold"
            if not task.child_task_set.all() and task.completed:
                peripheries = "2"
            graph.node(
                task.name,
                href=href,
                style=style,
                color="black",
                fontcolor="black",
                peripheries=peripheries,
            )
            for child in task.child_task_set.exclude(name="override"):
                graph.edge(task.name, child.name, color="black")

        for task in self.task_set.filter(name="override").prefetch_related(
            "parent_task_set", "child_task_set"
        ):
            label = "override_%s" % task.pk
            peripheries = "1"
            for parent in task.parent_task_set.all():
                graph.edge(parent.name, "override_%s" % task.pk, style="dashed")
            for child in task.child_task_set.all():
                graph.edge("override_%s" % task.pk, child.name, style="dashed")
            if not task.child_task_set.all() and task.completed:
                peripheries = "2"
            graph.node(label, style="filled, rounded, dashed", peripheries=peripheries)

        for task in self.task_set.exclude(name__in=names).exclude(name="override"):
            style = "filled, dashed"
            peripheries = "1"
            if task.type == tasks.HUMAN:
                style += ", rounded"
            if not task.completed:
                style += ", bold"
            for parent in task.parent_task_set.all():
                graph.edge(parent.name, task.name, style="dashed")
            for child in task.child_task_set.all():
                graph.edge(task.name, child.name, style="dashed")
            if not task.child_task_set.all() and task.completed:
                peripheries = "2"
            graph.node(
                task.name,
                style=style,
                color="black",
                fontcolor="black",
                peripheries=peripheries,
            )

        return graph

    def get_instance_graph_svg(self, output_format="svg"):
        """
        Return graph representation of a running workflow as SVG.

        The SVG is HTML safe and can be included in a template, e.g.:

        .. code-block:: html

            <html>
            <body>
            <!--// other content //-->
            {{ object.get_instance_graph_svg }}
            <!--// other content //-->
            </body>
            </html>

        Returns:
            (django.utils.safestring.SafeString): SVG representation of a running workflow.

        """
        graph = self.get_instance_graph()
        graph.format = output_format
        return SafeString(graph.pipe(encoding="utf-8"))  # nosec

    get_instance_graph_svg.short_description = t("instance graph")

    def cancel(self, user=None):
        self.task_set.cancel(user)


def workflow_state_subclasses():
    from django.apps import apps

    apps.check_models_ready()
    query = models.Q()
    for workflow in utils.get_workflows():
        opts = workflow._meta
        query |= models.Q(app_label=opts.app_label, model=opts.model_name)
    return query


class TasksQuerySet(models.query.QuerySet):
    def scheduled(self):
        return self.filter(status=self.model.SCHEDULED)

    def not_scheduled(self):
        return self.exclude(status=self.model.SCHEDULED)

    def succeeded(self):
        return self.filter(status=self.model.SUCCEEDED)

    def not_succeeded(self):
        return self.exclude(status=self.model.SUCCEEDED)

    def failed(self):
        return self.filter(status=self.model.FAILED)

    def canceled(self):
        return self.filter(status=self.model.CANCELED)

    def cancel(self, user=None):
        if user and not user.is_authenticated:
            user = None
        return self.update(
            status=self.model.CANCELED,
            completed_by_user=user,
            completed=Now(),
        )


class Task(models.Model):
    id = models.BigAutoField(primary_key=True, editable=False)
    _workflow = models.ForeignKey(
        Workflow,
        on_delete=models.CASCADE,
        db_column="workflow_id",
        editable=False,
    )
    content_type = models.ForeignKey(
        "contenttypes.ContentType",
        on_delete=models.CASCADE,
        editable=False,
        limit_choices_to=workflow_state_subclasses,
        related_name="joeflow_task_set",
    )
    workflow = GenericForeignKey(
        "content_type", "_workflow_id", for_concrete_model=False
    )

    name = models.CharField(max_length=255, db_index=True, editable=False)

    HUMAN = "human"
    MACHINE = "machine"
    _type_choices = (
        (HUMAN, t(HUMAN)),
        (MACHINE, t(MACHINE)),
    )
    type = models.CharField(
        max_length=50,
        choices=_type_choices,
        editable=False,
        db_index=True,
    )

    parent_task_set = models.ManyToManyField(
        "self",
        related_name="child_task_set",
        editable=False,
        symmetrical=False,
    )

    FAILED = "failed"
    SUCCEEDED = "succeeded"
    SCHEDULED = "scheduled"
    CANCELED = "canceled"
    _status_choices = (
        (FAILED, t(FAILED)),
        (SUCCEEDED, t(SUCCEEDED)),
        (SCHEDULED, t(SCHEDULED)),
        (CANCELED, t(CANCELED)),
    )
    status = models.CharField(
        max_length=50,
        choices=_status_choices,
        default=SCHEDULED,
        editable=False,
        db_index=True,
    )

    assignees = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        verbose_name=t("assignees"),
        related_name="joeflow_assignee_task_set",
    )

    completed_by_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=t("completed by"),
        related_name="joeflow_completed_by_task_set",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    created = models.DateTimeField(auto_now_add=True, db_index=True)
    modified = models.DateTimeField(auto_now=True, db_index=True)
    completed = models.DateTimeField(
        blank=True, null=True, editable=False, db_index=True
    )

    exception = models.TextField(blank=True)
    stacktrace = models.TextField(blank=True)

    objects = TasksQuerySet.as_manager()

    class Meta:
        ordering = ("-completed", "-created")
        get_latest_by = ("created",)
        permissions = (
            ("rerun", t("Can rerun failed tasks.")),
            ("cancel", t("Can cancel failed tasks.")),
        )
        default_manager_name = "objects"

    def __str__(self):
        return "%s (%s)" % (self.name, self.pk)

    def save(self, **kwargs):
        if self.pk:
            try:
                update_fields = kwargs["update_fields"]
            except KeyError as e:
                raise ValueError(
                    "You need to provide explicit 'update_fields' to avoid race conditions."
                ) from e
            else:
                update_fields.append("modified")
        super().save(**kwargs)

    def get_absolute_url(self):
        if self.completed:
            return
        url_name = "{}:{}".format(self.workflow.get_url_namespace(), self.name)
        try:
            return reverse(url_name, kwargs=dict(pk=self.pk))
        except NoReverseMatch:
            pass

    @property
    def node(self):
        return getattr(type(self.workflow), self.name)

    def finish(self, user=None):
        self.completed = timezone.now()
        self.status = self.SUCCEEDED
        if user and not user.is_authenticated:
            user = None
        self.completed_by_user = user
        if self.pk:
            self.save(update_fields=["status", "completed", "completed_by_user"])
        else:
            self.save()

    def cancel(self, user=None):
        self.completed = timezone.now()
        self.status = self.CANCELED
        if user and not user.is_authenticated:
            user = None
        self.completed_by_user = user
        self.save(update_fields=["status", "completed", "completed_by_user"])

    def fail(self):
        self.completed = timezone.now()
        self.status = self.FAILED
        tb = traceback.format_exception(*sys.exc_info())
        self.exception = tb[-1].strip()
        self.stacktrace = "".join(tb)
        self.save(update_fields=["status", "exception", "stacktrace"])

    def enqueue(self, countdown=None, eta=None):
        """
        Schedule the tasks for execution.

        Args:
            countdown (int):
                Time in seconds until the time should be started.

            eta (datetime.datetime):
                Time at which the task should be started.

        Returns:
            celery.result.AsyncResult: Celery task result.

        """
        self.status = self.SCHEDULED
        self.completed = None
        self.exception = ""
        self.stacktrace = ""
        self.save(update_fields=["status", "completed", "exception", "stacktrace"])
        task_runner = import_string(settings.JOEFLOW_TASK_RUNNER)
        transaction.on_commit(
            lambda: task_runner(
                task_pk=self.pk,
                workflow_pk=self._workflow_id,
                countdown=countdown,
                eta=eta,
            )
        )

    def start_next_tasks(self, next_nodes: list = None):
        """
        Start new tasks following another tasks.

        Args:
            self (Task): The task that precedes the next tasks.
            next_nodes (list):
                List of nodes that should be executed next. This argument is
                optional. If no nodes are provided it will default to all
                possible edges.

        """
        if next_nodes is None:
            next_nodes = self.workflow.get_next_nodes(self.node)
        tasks = []
        for node in next_nodes:
            try:
                # Some nodes – like Join – implement their own method to create new tasks.
                task = node.create_task(self.workflow, self)
            except AttributeError:
                task = self.workflow.task_set.create(
                    name=node.name, type=node.type, workflow=self.workflow
                )
            task.parent_task_set.add(self)
            if callable(node):
                transaction.on_commit(task.enqueue)
            tasks.append(task)
        return tasks



from django.contrib import admin, messages
from django.contrib.auth import get_permission_codename
from django.db import transaction
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as t

from . import forms, models
from .contrib.reversion import VersionAdmin

__all__ = ("WorkflowAdmin",)


def rerun(modeladmin, request, queryset):
    succeeded = queryset.succeeded().count()
    if succeeded:
        messages.warning(
            request,
            "Only failed tasks can be retried. %s tasks have been skipped" % succeeded,
        )
    counter = 0
    for obj in queryset.not_succeeded().iterator():
        obj.enqueue()
        counter += 1
    messages.success(request, "%s tasks have been successfully queued" % counter)


rerun.short_description = t("Rerun selected tasks")
rerun.allowed_permissions = ("rerun",)


def cancel(modeladmin, request, queryset):
    not_scheduled = queryset.not_scheduled().count()
    if not_scheduled:
        messages.warning(
            request,
            "Only scheduled tasks can be canceled. %s tasks have been skipped"
            % not_scheduled,
        )
    queryset.scheduled().cancel(request.user)
    messages.success(request, "Tasks have been successfully canceled")


cancel.short_description = t("Cancel selected tasks")
cancel.allowed_permissions = ("cancel",)


@admin.register(models.Task)
class TaskAdmin(VersionAdmin):
    def has_rerun_permission(self, request):
        opts = self.opts
        codename = get_permission_codename("rerun", opts)
        return request.user.has_perm("%s.%s" % (opts.app_label, codename))

    def has_cancel_permission(self, request):
        opts = self.opts
        codename = get_permission_codename("cancel", opts)
        return request.user.has_perm("%s.%s" % (opts.app_label, codename))

    def pretty_stacktrace(self, obj):
        return format_html('<pre class="readonly collapse">{}<pre>', obj.stacktrace)

    pretty_stacktrace.short_description = t("Traceback")

    def child_tasks(self, obj):
        return ", ".join(str(task) for task in obj.child_task_set.all().iterator())

    child_tasks.short_description = t("Child tasks")

    actions = (rerun, cancel)

    list_display = (
        "name",
        "status",
        "type",
        "content_type",
        "completed",
        "modified",
        "created",
    )

    readonly_fields = (
        "workflow",
        "name",
        "type",
        "parent_task_set",
        "child_tasks",
        "completed",
        "created",
        "modified",
        "exception",
        "pretty_stacktrace",
    )

    list_filter = (
        "status",
        "type",
        "content_type",
        "completed",
        "created",
    )

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "workflow",
                    "name",
                    "parent_task_set",
                    "child_tasks",
                    "completed",
                    "modified",
                    "created",
                    "exception",
                ),
            },
        ),
        (t("Traceback"), {"classes": ("collapse",), "fields": ("pretty_stacktrace",)}),
    )


class TaskInlineAdmin(admin.TabularInline):
    model = models.Task
    readonly_fields = [
        "name",
        "type",
        "assignees",
        "created",
        "completed",
        "completed_by_user",
        "status",
    ]
    fields = readonly_fields
    extra = 0
    can_delete = False
    show_change_link = True
    classes = ["collapse"]


class WorkflowAdmin(VersionAdmin):
    list_filter = (
        "modified",
        "created",
    )
    form = forms.OverrideForm

    def get_inlines(self, *args, **kwargs):
        return [*super().get_inlines(*args, **kwargs), TaskInlineAdmin]

    def get_readonly_fields(self, *args, **kwargs):
        return [
            "get_instance_graph_svg",
            *super().get_readonly_fields(*args, **kwargs),
            "modified",
            "created",
        ]

    @transaction.atomic()
    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        form.start_next_tasks(request.user)



# Generated by Django 3.1a1 on 2020-06-28 14:22

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import joeflow.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Workflow",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        editable=False, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
            ],
            options={"permissions": [("override", "Can override a workflow.")]},
        ),
        migrations.CreateModel(
            name="Task",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        editable=False, primary_key=True, serialize=False
                    ),
                ),
                (
                    "name",
                    models.CharField(max_length=255, db_index=True, editable=False),
                ),
                (
                    "type",
                    models.CharField(
                        max_length=50,
                        choices=[("human", "human"), ("machine", "machine")],
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        max_length=50,
                        choices=[
                            ("failed", "failed"),
                            ("succeeded", "succeeded"),
                            ("scheduled", "scheduled"),
                            ("canceled", "canceled"),
                        ],
                        db_index=True,
                        default="scheduled",
                        editable=False,
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("modified", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "completed",
                    models.DateTimeField(
                        blank=True, db_index=True, editable=False, null=True
                    ),
                ),
                ("exception", models.TextField(blank=True)),
                ("stacktrace", models.TextField(blank=True)),
                (
                    "_workflow",
                    models.ForeignKey(
                        db_column="workflow_id",
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="joeflow.workflow",
                    ),
                ),
                (
                    "assignees",
                    models.ManyToManyField(
                        related_name="joeflow_assignee_task_set",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="assignees",
                    ),
                ),
                (
                    "completed_by_user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="joeflow_completed_by_task_set",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="completed by",
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        editable=False,
                        limit_choices_to=joeflow.models.workflow_state_subclasses,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="joeflow_task_set",
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "parent_task_set",
                    models.ManyToManyField(
                        editable=False, related_name="child_task_set", to="joeflow.Task"
                    ),
                ),
            ],
            options={
                "ordering": ("-completed", "-created"),
                "permissions": (
                    ("rerun", "Can rerun failed tasks."),
                    ("cancel", "Can cancel failed tasks."),
                ),
                "get_latest_by": ("created",),
                "default_manager_name": "objects",
            },
        ),
    ]



import logging

from celery import shared_task
from django.apps import apps
from django.db import OperationalError, transaction

from joeflow.conf import settings
from joeflow.contrib.reversion import with_reversion

logger = logging.getLogger(__name__)


__all__ = ["task_runner"]


@shared_task(
    bind=True,
    ignore_results=True,
    max_retries=None,
    autoretry_for=(OperationalError,),
    retry_backoff=True,
)
def _celery_task_runner(self, task_pk, workflow_pk):
    Task = apps.get_model("joeflow", "Task")
    with transaction.atomic():
        task = Task.objects.select_for_update().get(pk=task_pk, completed=None)
        workflow = (
            task.content_type.model_class()
            .objects.select_for_update(nowait=True)
            .get(pk=workflow_pk)
        )

        try:
            logger.info("Executing %r", task)
            node = task.node
            with_task = getattr(node, "with_task", False)
            kwargs = {}
            if with_task:
                kwargs["task"] = task
            with with_reversion(task):
                result = node(workflow, **kwargs)
        except OperationalError:
            raise
        except:  # NoQA
            task.fail()
            logger.exception("Execution of %r failed", task)
        else:
            if result is False:
                logger.info("Task returned False, retrying …")
                transaction.on_commit(lambda: self.retry())
                return
            elif result is True:
                result = None
            logger.info("Task completed successful, starting next tasks: %s", result)
            task.start_next_tasks(next_nodes=result)
            task.finish()


def task_runner(*, task_pk, workflow_pk, countdown, eta):
    """Schedule asynchronous machine task using celery."""
    _celery_task_runner.apply_async(
        args=(task_pk, workflow_pk),
        countdown=countdown,
        eta=eta,
        queue=settings.JOEFLOW_CELERY_QUEUE_NAME,
    )



"""The lean workflow automation framework for machines with heart."""
import django

from . import _version

__version__ = _version.version
VERSION = _version.version_tuple

if django.VERSION < (3, 2):
    default_app_config = "joeflow.apps.JoeflowConfig"
